@@ -3,7 +3,7 @@
 
 import numpy as np
 from numba import jitclass, njit
-from numba import types
+from numba import types, _helperlib
 from numba.types import float32, boolean, uint32, string, void, int32
 
 from .checks import check_X_y, check_array
@@ -21,7 +21,6 @@ spec = [
     ("dirichlet", float32),
     ("split_pure", boolean),
     ("n_jobs", uint32),
-    ("random_state", int32),
     ("verbose", boolean),
     ("trees", types.List(TreeClassifier.class_type.instance_type, reflected=True)),
     ("samples", SamplesCollection.class_type.instance_type),
@@ -45,12 +44,10 @@ class AMFClassifierNoPython(object):
         dirichlet,
         split_pure,
         n_jobs,
-        random_state,
         verbose,
     ):
-
-        self.n_features = n_features
         self.n_classes = n_classes
+        self.n_features = n_features
         self.n_estimators = n_estimators
         self.step = step
         self.loss = loss
@@ -58,9 +55,7 @@ class AMFClassifierNoPython(object):
         self.dirichlet = dirichlet
         self.split_pure = split_pure
         self.n_jobs = n_jobs
-        self.random_state = random_state
         self.verbose = verbose
-
         self.iteration = 0
 
         samples = SamplesCollection()
@@ -68,36 +63,42 @@ class AMFClassifierNoPython(object):
 
         # TODO: reflected lists will be replaced by typed list soon...
         trees = [
-            TreeClassifier(n_features, n_classes, samples) for _ in range(n_estimators)
+            TreeClassifier(
+                self.n_classes,
+                self.n_features,
+                self.step,
+                self.loss,
+                self.use_aggregation,
+                self.dirichlet,
+                self.split_pure,
+                self.samples,
+            )
+            for _ in range(n_estimators)
         ]
         self.trees = trees
 
-    # TODO: put it outside as a @njit function
-    def partial_fit(self, X, y):
-        n_samples_batch, n_features = X.shape
-        # We need at least the actual number of nodes + twice the extra samples
 
-        # Let's reserve nodes in advance
-        # print('Reserve nodes in advance')
-        for tree in self.trees:
-            n_nodes_reserved = tree.nodes.n_nodes + 2 * n_samples_batch
-            # print('n_nodes_reserved:', n_nodes_reserved)
-            tree.nodes.reserve_nodes(n_nodes_reserved)
-        # print('Done.')
+@njit(
+    void(AMFClassifierNoPython.class_type.instance_type, float32[:, ::1], float32[::1])
+)
+def partial_fit(forest, X, y):
+    n_samples_batch, n_features = X.shape
+    # We need at least the actual number of nodes + twice the extra samples
 
-        # First, we save the new batch of data
-        # print('Append the new batch of data')
-        n_samples_before = self.samples.n_samples
-        self.samples.append(X, y)
-        # print('Done.')
+    # Let's reserve nodes in advance
+    for tree in forest.trees:
+        n_nodes_reserved = tree.nodes.n_nodes + 2 * n_samples_batch
+        tree.nodes.reserve_nodes(n_nodes_reserved)
 
-        for i in range(n_samples_before, n_samples_before + n_samples_batch):
-            # Then we fit all the trees using all new samples
-            for tree in self.trees:
-                # print('i:', i)
-                tree_partial_fit(tree, i)
+    # First, we save the new batch of data
+    n_samples_before = forest.samples.n_samples
+    forest.samples.append(X, y)
 
-            self.iteration += 1
+    for i in range(n_samples_before, n_samples_before + n_samples_batch):
+        # Then we fit all the trees using all new samples
+        for tree in forest.trees:
+            tree_partial_fit(tree, i)
+        forest.iteration += 1
 
     # def predict(self, X, scores):
     #     scores.fill(0.0)
@@ -259,7 +260,7 @@ class AMFClassifier(object):
         self._classes = set(range(n_classes))
 
     def partial_fit(self, X, y):
-        n_samples, n_features = X.shape
+        # TODO: write the docstring
         # First,ensure that X and y are C-contiguous and with float32 dtype
         X, y = check_X_y(
             X,
@@ -278,6 +279,7 @@ class AMFClassifier(object):
             y_numeric=True,
             estimator="AMFClassifier",
         )
+        n_samples, n_features = X.shape
 
         if y.min() < 0:
             raise ValueError("All the values in `y` must be non-negative")
@@ -299,7 +301,6 @@ class AMFClassifier(object):
                 self.dirichlet,
                 self.split_pure,
                 self.n_jobs,
-                self._random_state,
                 self.verbose,
             )
         else:
@@ -309,8 +310,9 @@ class AMFClassifier(object):
                     "`partial_fit` was first called with n_features=%d while "
                     "n_features=%d in this call" % (self.n_features, n_features)
                 )
-
-        self.no_python.partial_fit(X, y)
+        self._set_random_state()
+        partial_fit(self.no_python, X, y)
+        self._put_back_random_state()
         return self
 
     def predict_proba(self, X):
@@ -318,13 +320,13 @@ class AMFClassifier(object):
 
         Parameters
         ----------
-        X : `np.ndarray` or `scipy.sparse.csr_matrix`, shape=(n_samples, n_features)
-            Features matrix to predict for.
+        X : `np.ndarray`, shape=(n_samples, n_features)
+            Features matrix to predict for
 
         Returns
         -------
         output : `np.ndarray`, shape=(n_samples, n_classes)
-            Returns predicted values.
+            Returns the predicted class probabilities
         """
         import numpy as np
 
@@ -355,7 +357,9 @@ class AMFClassifier(object):
                     "`partial_fit` was called with n_features=%d while `predict_proba` "
                     "received n_features=%d" % (self.n_features, n_features)
                 )
+        self._set_random_state()
         predict_proba(self.no_python, X, scores)
+        self._put_back_random_state()
         return scores
 
     # def predict(self, X):
@@ -366,6 +370,26 @@ class AMFClassifier(object):
     #         scores = self.predict_proba(X)
     #         return scores.argmax(axis=1)
 
+    def _set_random_state(self):
+        # This uses a trick by Alexandre Gramfort,
+        #   see https://github.com/numba/numba/issues/3249
+        if self._random_state >= 0:
+            r = np.random.RandomState(self._random_state)
+            ptr = _helperlib.rnd_get_np_state_ptr()
+            ints, index = r.get_state()[1:3]
+            _helperlib.rnd_set_state(ptr, (index, [int(x) for x in ints]))
+            self._ptr = ptr
+            self._r = r
+
+    def _put_back_random_state(self):
+        # This uses a trick by Alexandre Gramfort,
+        #   see https://github.com/numba/numba/issues/3249
+        if self._random_state >= 0:
+            ptr = self._ptr
+            r = self._r
+            index, ints = _helperlib.rnd_get_state(ptr)
+            r.set_state(("MT19937", ints, index, 0, 0.0))
+
     def get_nodes_df(self, idx_tree):
         import pandas as pd
 
@@ -424,7 +448,6 @@ class AMFClassifier(object):
         df = pd.DataFrame(data, columns=columns)
         return df
 
-    # TODO: code properties for everything
     @property
     def n_classes(self):
         return self._n_classes
